<?php
if (!defined('MEDIAWIKI')) die("MediaWiki extensions cannot be run directly.");
if (!defined('WikiDB')) {
	die("The WikiDB extension has not been set up correctly.<br>'"
		. basename(__FILE__) . "' should not be included directly.<br>Instead, "
		. "add the following to LocalSettings.php: "
		. "<tt><b>include_once(\"WikiDB/WikiDB.php\");</b></tt>");
}
/**
 * Class to handle queries of the WikiDB.
 *
 * @package MediaWiki
 * @subpackage Extensions
 *
 * @author Mark Clements <mclements at kennel17 dot co dot uk>
 * @copyright Copyright  2006-2010, Mark Clements
 * @license http://creativecommons.org/licenses/by-sa/2.5/ cc-by-sa 2.5 or later
 * @version $Rev: 475 $
 */

WikiDB_RegisterRevision('$Rev: 475 $');

class WikiDB_Query {
	var $pDB;
	var $pErrorMsg;

	var $pTables = array();
	var $pSortFields = array();
	var $pCriteriaFields = array();

/////////////// PUBLIC FUNCTIONS ///////////////

// WikiDB_Query() - Constructor
// Takes a table string, and optional criteria and sort strings, which are used to
// set up the query object ready for running.
// When the object has been initialised, call the Run() function to perform the 
// actual query and return the full set of results.
	function WikiDB_Query($TableString, $CriteriaString = "", $SortString = "") {
		$this->pDB =& wfGetDB(DB_SLAVE);
		$this->pErrorMsg = "";

		$this->pSetTablesFromString($TableString);
		$this->pSetCriteriaFromString($CriteriaString);
		$this->pSetSortFromString($SortString);
	}

// Run()
// Runs the defined query and returns an array containing the results, one item per
// row, indexed numerically.
// If an error occurs, a string is returned containing the error message.
	function Run() {
		if ($this->pErrorMsg != "")
			return $this->pErrorMsg;

		$this->pGenerateFieldLists();
		$this->pCreateTempTables();
		return $this->pPerformQuery();
	}

/////////////// SETTING UP INPUTS ///////////////

// pSetTablesFromString()
// Takes a table string and sets up the array of tables used in this query.
// Currently a table string may only contain a single table.  In future we may
// allow multiple tables, or joins.
// Table aliases are resolved before we add the data to our array.
// The tables array is indexed by the full table name, and each element contains the
// necessary data pertaining to the given table (the table object, the name of the
// temporary table that we will build, and an array of referenced fields, initially
// empty).
// TODO: Implement multi-table queries.
	function pSetTablesFromString($TableString) {
	// Create a table object for this table, resolving any redirects.
		$Table = new WikiDB_Table($TableString);
		$Table = $Table->GetDataDestination();

	// Get canonical name of table
		$TableName = $Table->GetFullName();

	// Create a name for the resulting temporary table.
	// We use a random string, as there may be several instances of the same table
	// on one page.
	// TODO: Once multiple tables are allowed we will need to ensure this remains
	//		 within the current request.
		$TempName = md5(rand());

		$this->pTables[$TableName]['table'] = $Table;
		$this->pTables[$TableName]['tempname'] = $TempName;
		$this->pTables[$TableName]['fields'] = array();
	}

// pSetCriteriaFromString()
// Parses the criteria string, and sets up an array of criteria data indicating
// the involved fields and the operator(s) that apply.
// This function is currently very crude, because the allowed syntax is very
// crude - this all needs expanding.
// TODO: Allow for multiple tables.
// TODO: Expand, expand, expand!
//
// The current criteria string may contain the following:
// 	* Multiple criteria are allowed, separated by commas.  These will be ANDed
//	  together when the query is run (currently no way to OR criteria).
//	* Each criteria is of the form "FIELD OPERATOR VALUE", e.g. "Artist=Madonna".
//	* Whitespace is allowed in the criteria string, and is automatically removed.
//
//	* FIELD is of the form "TABLE.FIELDNAME", where "TABLE." is optional if there is
//	  only one table being used in the query.  If more than one table is being used,
//	  then it is required for all fields (not applicable currently).
//  * TABLE aliases are resolved properly.
//	* The resulting TABLE (after alias resolution) must match one of the tables 
//    defined for the query.
//	* FIELDNAME does not have to be defined in the table, as undefined fields may
//    still contain data.
//	* Field aliases are resolved properly if FIELDNAME exists in the table 
//	  definition.
//
//	* OPERATOR is any of the valid operators listed in $ValidOperators.
//
//	* VALUE is the value for comparison, unquoted.  Currently fields cannot be
// 	  compared to each other.
//
	function pSetCriteriaFromString($CriteriaString) {
		$ValidOperators = array("<>", ">=", "<=", "!=", ">", "<", "=");

	// Setup regular expression for matching individual criteria.
		$Regex = "";
		foreach ($ValidOperators as $Operator) {
			if (!IsBlank($Regex))
				$Regex .= "|";
			$Regex .= preg_quote($Operator);
		}
		$Regex = '/^(.*)((?-U)' . $Regex . ')(.*)$/iU';

		$WhereFields = array();

//		print("-> " . $CriteriaString . "<br>");

	// Criteria are comma-separated, so explode on commas to get the individual criteria.
		foreach (explode(",", $CriteriaString) as $Criteria) {
//			print("* " . $Criteria . "<br>");
			if(preg_match($Regex, $Criteria, $Matches)) {
//				print_r($Matches);
				$Field = trim($Matches[1]);
				$Value = trim($Matches[3]);
				$Operator = $Matches[2];

			// Modify the != operator to use the <> operator, as the former is not
			// recognised by MySQL.
				if ($Operator == "!=")
					$Operator = "<>";

			// Parse the field to get the proper table/field names.
			// All aliases are resolved.
				$FieldData = $this->pParseFieldName($Field);

			// If an error occurred, then return without doing anything more.
				if ($FieldData == false)
					return;

				$Value = $FieldData['table']->FormatFieldForSorting(
														$FieldData['fieldname'],
														$Value
													);

			// The Value parameter is required, so if it is not set ignore this
			// criterion.
				if ($Value != "") {
					$WhereFields[] = array(
											'table' => $FieldData['tablename'],
											'field' => $FieldData['fieldname'],
											'op' => $Operator,
											'value' => $Value,
										);
				}
			}
		}

		$this->pCriteriaFields = $WhereFields;
	}

// pSetSortFromString()
// Parses the sort string, and sets up an array of sort data indicating the fields
// to sort on and the direction of sort.
// TODO: Allow for multiple tables.
//
// The sort string may contain the following:
//	* Multiple sort criteria are allowed, separated by commas.  The resulting data
//	  will be sorted by each of these fields in order, in the same way as a standard
//	  ORDER BY clause.
//	* Each criterion is of the form "FIELD DIRECTION", where DIRECTION is optional.
//	* Whitespace is allowed and ignored.
//
//	* FIELD follows the same rules as for the WHERE criteria, above.
//	* DIRECTION is either "ASC" or "DESC" (case-insensitive).  If omitted, "ASC"
//    is used.
//
	function pSetSortFromString($SortString) {
		$SortFields = array();

		foreach (explode(",", $SortString) as $SortField) {
			if(preg_match('/^(.*)(?:\s(ASC|DESC))?\s*$/iU', $SortField, $Matches)) {
				$Field = trim($Matches[1]);
				if ($Field == "")
					continue;

			// Normalise the $Direction property, setting it if omitted.
			// We don't need to check the value is valid, as the regex has already
			// done that for us.
				if (!isset($Matches[2]))
					$Direction = "ASC";
				else
					$Direction = strtoupper($Matches[2]);

			// Parse the field to get the proper table/field names.
			// All aliases are resolved.
				$FieldData = $this->pParseFieldName($Field);

			// If an error occurred, then return without doing anything more.
				if ($FieldData == false)
					return;

				$SortFields[] = array(
									'table' => $FieldData['tablename'],
									'field' => $FieldData['fieldname'],
									'dir' => $Direction,
								// We also set 'key' for compatibility with 
								// the ArraySorter class.
								// TODO: Perhaps just switch to this index name,
								//		 rather than duplicating 'field'?
									'key' => $Field,
								);
			}
		}

		$this->pSortFields = $SortFields;
	}

// pParseFieldName()
// Parses a string to get the field name.  If the field string contains a period
// then the portion prior to the period is the name of a table, otherwise the table
// defined for the query is assumed (and if more than one is defined, an error is
// raised).
// If the string contains a table component and it wasn't in the query's table
// string then an error is raised.
// If valid, an array containing the 'tablename' and 'fieldname' is returned.  The
// array also contains the WikiDB_Table object for the table, to save calling 
// functions having to recreate it, if they need access to it for any reason.
// Otherwise pError is set and false is returned.
// Table aliases are resolved, and so are field aliases, if the field exists in 
// the table definition.
	function pParseFieldName($FieldString) {
	// Get position of period in field name
		$DotPos = strpos($FieldString, ".");

	// If there is no period in the field string, then use the table defined in the
	// query's table string.
		if ($DotPos === false) {
		// If there is more than one table defined for this query, then the field
		// string is ambiguous, so raise an error.
			if (count($this->pTables) > 1) {
				$this->pError = "Ambiguous fields in WHERE clause.";
				return false;
			}

		// Otherwise, use the table defined in the query's table string.
		// This has already had aliases resolved.
			$FullTableName = key($this->pTables);
			$Table = $this->pTables[$FullTableName]['table'];
		}

	// If the field string contains a period then it contains a table name
	// component, which needs separating and checking.
		else {
		// Extract table name from Field
			$TableName = trim(substr($FieldString, 0, $DotPos - 1 ));
		// Set $FieldString to hold just the field component.
			$FieldString = trim(substr($FieldString, $DotPos + 1));

		// Resolve table aliases, and resolve table name to it's canonical version
		// (i.e. including namespace).
			$Table = new WikiDB_Table($TableName);
			$Table = $Table->GetDataDestination();
			$FullTableName = $Table->GetFullName();

		// If the table is not defined for use in the query, raise an error.
			if (!isset($this->pTables[$FullTableName])) {
				$this->pError = "Undefined table '" . $TableName
							  . "' in WHERE clause.";
				return false;
			}
		}

	// Resolve field aliases if the field is defined.
		if ($Table->FieldIsDefined($FieldString))
			$FieldString = $Table->GetAlias($FieldString);

		return array(
					'tablename' => $FullTableName, 
					'fieldname' => $FieldString,
					'table'		=> $Table,
				);
	}

/////////////// GENERATING INTERMEDIATE DATA ///////////////

// pGenerateFieldLists()
// Fills in the 'fields' array for each table.  This will become an array of fields 
// in that table which are referenced in the query for filtering or sorting purposes
// (i.e. the fields that we need to include in the output).  
// The array is indexed by the actual field name, with the value being the name
// we will use for the table when creating the join statement required for that 
// field.
// All table and field aliases have been resolved prior to calling this function.
	function pGenerateFieldLists() {
		$FieldNames = array();

		if (count($this->pSortFields) > 0) {
			foreach ($this->pSortFields as $SortField)
				$FieldNames[$SortField['table']][] = $SortField['field'];
		}
		if (count($this->pCriteriaFields) > 0) {
			foreach ($this->pCriteriaFields as $WhereField)
				$FieldNames[$WhereField['table']][] = $WhereField['field'];
		}

		foreach ($FieldNames as $TableName => $Fields) {
			array_unique($Fields);

			foreach ($Fields as $FieldName) {
				$FieldTableName = QuoteIdentifier("wikidbtmp_" . $FieldName);
				$this->pTables[$TableName]['fields'][$FieldName] = $FieldTableName;
			}
		}
	}

// pCreateTempTables()
// Creates the individual temporary tables which we will subsequently query.
	function pCreateTempTables() {
		foreach ($this->pTables as $TableName => $TableData) {
			$SQL = $this->pGetSubSelectForTable(
										$TableData['table'],
										$TableData['fields']
									);
			
			$SQL = "CREATE TEMPORARY TABLE " 
				 . $this->pDB->tableName($TableData['tempname'])
				 . " " . $SQL;

			$Result = $this->pDB->query($SQL);
		}
	}

// pGetSubSelectForTable()
// Returns a select query that builds a virtualisation of the specified WikiDB
// table, including just the columns specified in the field list, plus the standard
// row data in wikidb_rowdata.
// Note that when stepping through the field lists, the $FieldName is not yet
// escaped, and so always needs quoting, whereas the $FieldTableName has already
// been escaped properly.  The quoting for $FieldName differs as in some places it
// is used as a column name, where in others it is a string to be matched.
	function pGetSubSelectForTable($Table, $FieldList) {
	// Create a short-cut, for tidier code.
		$DB =& $this->pDB;

	// Get the correct table names for the real tables (properly escaped).
		$RowTable = $DB->tableName(pWIKIDB_tblRows);
		$FieldTable = $DB->tableName(pWIKIDB_tblFields);

	// First column is the parsed data for the row.
		$SQL = "SELECT " 
		     . $RowTable . ".page_namespace AS _page_namespace, "
			 . $RowTable . ".page_title AS _page_title, "
			 . $RowTable . ".parsed_data AS _parsed_data";

	// Then add each of the requested fields.
		if (count($FieldList) > 0) {
			foreach ($FieldList as $FieldName => $FieldNameTable) {
				$SQL .= ", " . $FieldNameTable . ".field_value AS "
					  . QuoteIdentifier($FieldName);
			}
		}

	// Create JOIN expression.  We join the row table to multiple instances of the
	// field table (one per field).
		$SQL .= " FROM " . $RowTable;
		if (count($FieldList) > 0) {
			foreach ($FieldList as $FieldName => $FieldNameTable) {
				$FieldString = "";
				$Aliases = $Table->GetAliasRev($FieldName);
		 
				foreach ($Aliases as $Alias) {
					if (!IsBlank($FieldString))
						$FieldString .= " OR ";
					$FieldString .= $FieldNameTable . ".field_name = "
								  . $DB->addQuotes($Alias);
				}

				$SQL .= " LEFT JOIN " . $FieldTable . " AS " . $FieldNameTable
					  . " ON " . $FieldNameTable . ".row_id = "
					  . $RowTable . ".row_id" . " AND (" . $FieldString . ")";
			}
		}

	// Create WHERE clause to restrict to the required table.
		$SQL .= " WHERE ";
		foreach ($Table->GetTableAliases() as $Index => $TableAlias) {
			if ($Index > 0)
				$SQL .= " OR ";

			$SQL .= "(" . $RowTable . ".table_namespace = "
				  . $DB->addQuotes($TableAlias->GetNamespace())
				  . " AND " . $RowTable . ".table_title = "
				  . $DB->addQuotes($TableAlias->GetDBKey())
				  . ")";
		}

		$SQL .= ";";

		return $SQL;
	}

/////////////// PERFORMING THE QUERY ///////////////

// pPerformQuery()
// Performs the actual query, using the temporary tables already created.
// TODO: Allow multi-table queries.
	function pPerformQuery() {
	// Block multiple tables for now.
		if (count($this->pTables) != 1) {
			$this->pErrorMsg = "Currently only queries involving a single table "
							 . "are supported.";
			return;
		}

		reset($this->pTables);
		$TableData = $this->pTables[key($this->pTables)];
		$Query = $this->pGetQuerySQL($TableData, $this->pCriteriaFields,
									 $this->pSortFields);
		$Result = $this->pDB->query($Query);

//		print_r(htmlspecialchars($TempTable) . "<br>\n" . htmlspecialchars($Query)
//				. "<br>\n");

		$RowData = array();
		$RowID = 0;
		while ($Row = $this->pDB->fetchObject($Result)) {
		// Extract the data for this row.
			$RowData[$RowID]['Data'] = unserialize($Row->_parsed_data);
		
		// Add the meta-data.
			$RowData[$RowID]['MetaData']['_Row'] = $RowID + 1;
			$RowData[$RowID]['MetaData']['_SourceArticle'] = 
								WikiDB_Table::GetFullName($Row->_page_title, 
														  $Row->_page_namespace);

			$RowID++;
		}
		$this->pDB->freeResult($Result);

//		print_r($RowData);
		return $RowData;
	}

// Construct query to retrieve data from the temporary table.
// TODO - Not currently safe from SQL injection!
	function pGetQuerySQL($TableData, $WhereFields, $SortFields) {
	// Get the correct table names for the real tables (properly escaped).
		$RowTable = $this->pDB->tableName(pWIKIDB_tblRows);
		$FieldTable = $this->pDB->tableName(pWIKIDB_tblFields);

	// Escape temporary table name
		$TempTableName = $this->pDB->tableName($TableData['tempname']);

	// Select all rows
		$SQL = "SELECT * FROM " . $TempTableName;

	// Add WHERE clause if there is one
		if (count($WhereFields) > 0) {
			$SQL .= " WHERE ";
			foreach ($WhereFields as $i => $Criteria) {
				if ($i > 0)
					$SQL .= " AND ";
				$SQL .= QuoteIdentifier($Criteria['field'])
					  . " " . $Criteria['op'] . " "
					  . $this->pDB->addQuotes($Criteria['value']);
			}
		}

	// Add ORDER BY clause if there is one
	// TODO: Not safe from SQL injection!
		if (count($SortFields) > 0) {
			$SQL .= " ORDER BY ";
			foreach ($SortFields as $i => $Sort) {
				if ($i > 0)
					$SQL .= ", ";
				$SQL .= QuoteIdentifier($Sort['field']);
				if ($Sort['dir'] == "DESC")
					$SQL .= " DESC";
			}
		}

		$SQL .= ";";

		return $SQL;
	}

} // END: class WikiDB_Query

